import logging
import argparse
import sys

from .cache import (
    get_response_words,
    delete_word,
)
from .console import c_print
from .log import logger
from .utils import OP, DICT, is_tool
from .dicts import webster, cambridge, dicts
from .__init__ import __version__

def parse_args():
    parser = argparse.ArgumentParser(
        description="Terminal Version of Cambridge Dictionary by default. Also supports Merriam-Webster Dictionary."
    )

    parser.add_argument(
        "-v",
        "--version",
        action="store_true",
        help="print the current version of the program",
    )

    # Add sub-command capability that can identify different sub-command name
    sub_parsers = parser.add_subparsers(dest="subparser_name")

    # Add sub-command l
    parser_lw = sub_parsers.add_parser(
        "l",
        help="list alphabetically ordered words/phrases you've found before",
    )

    # Make sub-command l run default funtion of "list_words"
    parser_lw.set_defaults(func=list_words)

    # Add an optional argument for l command
    parser_lw.add_argument(
        "-d",
        "--delete",
        nargs="+",
        help="delete a word/phrase or multiple words/phrases(separated by ', ') from cache",
    )

    # Add an optional argument for l command
    parser_lw.add_argument(
        "-t",
        "--time",
        action="store_true",
        help="list words/phrases you've found before in reverse chronological order",
    )

    # Add an optional argument for l command
    parser_lw.add_argument(
        "-r",
        "--random",
        action="store_true",
        help="randomly list 20 words/phrases you've found before",
    )

    # Add sub-command s
    parser_sw = sub_parsers.add_parser("s", help="look up a word/phrase; hidden for convenience, no need to type")

    # Make sub-command s run default function of "search_words"
    parser_sw.set_defaults(func=search_word)

    # Add positional arguments with n args for s command
    parser_sw.add_argument(
        "word_or_phrase",
        nargs="+",
        help="look up a word/phrase in Cambridge Dictionary; e.g. camb <word/phrase>",
    )

    # Add an optional argument for s command
    parser_sw.add_argument(
        "-d",
        "--debug",
        action="store_true",
        help="look up a word/phrase in debug mode",
    )

    # Add an optional argument for s command
    parser_sw.add_argument(
        "-w",
        "--webster",
        action="store_true",
        help="look up a word/phrase in Merriam-Webster Dictionary",
    )

    # Add an optional argument for s command
    parser_sw.add_argument(
        "-f",
        "--fresh",
        action="store_true",
        help="look up a word/phrase afresh without using cache",
    )

    # Add an optional argument for s command
    parser_sw.add_argument(
        "-c",
        "--chinese",
        action="store_true",
        help="look up a word/phrase in Cambridge Dictionary with Chinese translation",
    )

    # Add an optional argument for s command
    parser_sw.add_argument(
        "-n",
        "--nosuggestions",
        action="store_true",
        help="look up a word/phrase without showing spelling suggestions if not found",
    )

    # Add sub-command wod
    parser_wod = sub_parsers.add_parser(
        "wod",
        help="list today's Word of the Day from Merriam-Webster Dictionary",
    )

    # Make sub-command wod run default funtion of "wod"
    parser_wod.set_defaults(func=wod)

    # Add an optional argument for wod command
    parser_wod.add_argument(
        "-l",
        "--list",
        action="store_true",
        help="list all words of the day",
    )

    if len(sys.argv) == 1:
        print_help(parser, parser_lw, parser_sw, parser_wod)
        sys.exit()

    elif sys.argv[1] == "-h" or sys.argv[1] == "--help":
        print_help(parser, parser_lw, parser_sw, parser_wod)
        sys.exit()

    elif sys.argv[1] == "-v" or sys.argv[1] == "--version":
        print("cambridge " + __version__)
        sys.exit()

    elif sys.argv[1] == "l" or sys.argv[1] == "wod":
        args = parser.parse_args()
        return args

    else: # only for command "s"
        to_parse = []
        word = []
        for i in sys.argv[1 : ]:
            if i.startswith("-") and i[1] in ["-", "h", "d", "f", "w", "c", "n"]:
                to_parse.append(i)
            # NOTE
            # Python stdlib `parser.parse_args()` does not allow the value of an argument to start with "-", e.g. camb -w "-let"
            # Without the following workaround, Python interpreter will terminate the program  with the error:
            # `main.py s: error: the following arguments are required: word_or_phrase`
            # However, in terms of word lookup, you sometimes just want to know the meaning of a specific suffix like "-let"
            # Luckily, search without "-", the result of which can also contain the meaning of "-let" suffix
            elif i.startswith("-"):
                word.append(i[1 : ])
            else:
                word.append(i)

        to_search = " ".join(word)
        to_parse.append(to_search)
        args= parser_sw.parse_args(to_parse)
        return args


def print_help(parser, parser_lw, parser_sw, parser_wod):
    parser.print_help()
    print("\n\n\033[1mCOMMAND l\033[0m")
    parser_lw.print_help()

    print("\n\n\033[1mCOMMAND s (hidden)\033[0m")
    parser_sw.print_help()

    print("\n\n\033[1mCOMMAND wod\033[0m")
    parser_wod.print_help()

    sys.exit()


def delete(word):
    deleted, data = delete_word(word)

    if deleted and data is not None:
        for i in data:
            if "cambridge" in i[1]:
                print(f'{OP.DELETED.name} "{word}" from {DICT.CAMBRIDGE.name} in cache successfully')
            else:
                print(f'{OP.DELETED.name} "{word}" from {DICT.MERRIAM_WEBSTER.name} in cache successfully')
    else:
        print(f'{OP.NOT_FOUND.name} "{word}" in cache')


def list_words(args):
    # The subparser i.e. the sub-command isn't in the namespace of args

    if args.delete:
        to_delete = args.delete
        words = " ".join(to_delete)
        for w in words.split(","):
            i = w.strip()
            if i:
                delete(i)

    elif args.random:
        result = get_response_words(is_random=True)
        if not result[0]:
            logger.error("You may haven't searched any word yet")
        else:
            data = result[1]
            list_notice = "Select to print the item's meaning; [ESC] to quit out."
            if not is_tool("fzf"):
                print()
                dicts.list_items(data, "cache_list")
                print()
            else:
                dicts.list_items_fzf(data, "cache_list", list_notice, None, None, False)

    else:
        result = get_response_words(is_random=False)
        if not result[0]:
            logger.error("You may haven't searched any word yet")

        else:
            data = result[1]
            list_notice = "Select to print the item's meaning; [ESC] to quit out."
            if args.time:
                if not is_tool("fzf"):
                    data.sort(reverse=False, key=lambda tup: tup[2])
                    print()
                    dicts.list_items(data, "cache_list")
                    print()
                else:
                    data.sort(reverse=True, key=lambda tup: tup[2])
                    dicts.list_items_fzf(data, "cache_list", list_notice, None, None, False)
            else:
                data.sort()
                if not is_tool("fzf"):
                    print()
                    dicts.list_items(data, "cache_list")
                    print()
                else:
                    dicts.list_items_fzf(data, "cache_list", list_notice, None, None, False)


def search_word(args):
    """
    The function is triggered when a user searches a word or phrase on terminal.
    First checks the args having "verbose" in it or not, if so, the debug mode will be turned on.
    Then it checks which dictionary is intended, and then calls respective dictionary function.
    """

    if args.debug:
        logging.getLogger(__package__).setLevel(logging.DEBUG)


    input_word = args.word_or_phrase[0].strip(".").strip(",").strip()

    if not input_word:
        print("You didn't input any word or phrase.")
        sys.exit()

    # boolean types
    is_webster = args.webster
    is_fresh = args.fresh
    is_ch = args.chinese
    no_suggestions = args.nosuggestions

    if is_webster and is_ch:
        print("Webster Dictionary doesn't support English to other language. Try again without -c(--chinese) option")
        sys.exit()

    if is_webster:
        webster.search_webster(input_word, is_fresh, no_suggestions, None)
    else:
        cambridge.search_cambridge(input_word, is_fresh, is_ch, no_suggestions, None)


def wod(args):
    if args.list:
        webster.get_wod_list()

    else:
        webster.get_wod()



from pathlib import Path
import datetime
import sqlite3

dir = Path.home() / ".cache" / "cambridge"
dir.mkdir(parents=True, exist_ok=True)
DB = str(dir / "cambridge.db")

current_datetime = datetime.datetime.now()

con = sqlite3.connect(DB, detect_types=sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES)

def create_table():
    with con:
        con.execute(
            """CREATE TABLE words (
            "input_word" TEXT NOT NULL,
            "response_word" TEXT NOT NULL,
            "created_at" TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
            "response_url" TEXT UNIQUE NOT NULL,
            "response_text" TEXT NOT NULL,
            UNIQUE(response_url))"""
        )


def insert_into_table(input_word, response_word, url, text):
    response_word = response_word.lower()
    with con:
        try:
            con.execute(
                "INSERT INTO words (input_word, response_word, created_at, response_url, response_text) VALUES (?, ?, ?, ?, ?)",
                (input_word, response_word, current_datetime, url, text)
            )
        except sqlite3.IntegrityError as error:
            return (False, error)

        except sqlite3.InterfaceError as error:
            return (False, error)
        else:
            return (True, None)


def get_cache(word, request_url):
    try:
        with con:
            cur = con.execute(
                "SELECT response_url, response_word, response_text FROM words WHERE response_url = ? OR response_word = ? OR input_word = ?",
                (request_url, word, word)
            )
    except sqlite3.OperationalError:
        create_table()
    else:
        data = cur.fetchone()
        return data


def get_response_words(is_random=False):
    try:
        with con:
            if is_random:
                cur = con.execute("SELECT response_word, response_url FROM words ORDER BY RANDOM() LIMIT 20")
            else:
                cur = con.execute("SELECT response_word, response_url, created_at FROM words")
    except sqlite3.OperationalError:
        return (False, None)
    else:
        data = cur.fetchall()
        return (True, data)


def delete_word(word):
    with con:
        cur = con.execute("SELECT input_word, response_url FROM words WHERE input_word = ? OR response_word = ?", (word, word))
        data = cur.fetchall()

        if data == []:
            return (False, None)
        else:
            con.execute("DELETE FROM words WHERE input_word = ? OR response_word = ?", (word, word))
            return (True, data)



import sys
import os
import requests
import subprocess
from fake_user_agent import user_agent

from ..cache import insert_into_table, get_cache, delete_word
from ..log import logger
from ..errors import call_on_error
from ..dicts import cambridge, webster
from ..utils import make_a_soup, OP, DICT, is_tool
from ..console import c_print

from typing import (
    Optional,
    Literal
)

Initiator = Literal["wod_calendar", "spell_check", "cache_list", "redirect_list"]

def fetch(url, session):
    ua = user_agent()
    headers = {"User-Agent": ua}
    session.headers.update(headers)
    attempt = 0

    logger.debug(f"{OP.FETCHING.name} {url}")
    while True:
        try:
            r = session.get(url, timeout=9.05)
            # Only when calling raise_for_status, will requests raise HTTPError if any.
            # See: https://blog.csdn.net/Odaokai/article/details/100133503
            # for webster, only when status code is 404, can we know to redirect to spellcheck page, so you can't exit on 404
            if r.status_code >= 500:
                r.raise_for_status()

        except requests.exceptions.HTTPError as e:
            attempt = call_on_error(e, url, attempt, OP.RETRY_FETCHING.name)
            continue
        except requests.exceptions.ConnectTimeout as e:
            attempt = call_on_error(e, url, attempt, OP.RETRY_FETCHING.name)
            continue
        except requests.exceptions.ConnectionError as e:
            attempt = call_on_error(e, url, attempt, OP.RETRY_FETCHING.name)
            continue
        except requests.exceptions.ReadTimeout as e:
            attempt = call_on_error(e, url, attempt, OP.RETRY_FETCHING.name)
            continue
        except Exception as e:
            attempt = call_on_error(e, url, attempt, OP.RETRY_FETCHING.name)
            continue
        else:
            return r


def cache_run(input_word, req_url):
    # data is a tuple (response_url, response_text) if any
    data = get_cache(input_word, req_url)

    if data is None:
        if "s" != input_word[-1]:
            return False
        else:
            data = get_cache(input_word[:-1], req_url)
            if data is None:
                if "es" != input_word[-2:]:
                    return False
                else:
                    data = get_cache(input_word[:-2], req_url)
                    if data is None:
                        return False

    res_url, res_word, res_text = data

    if DICT.CAMBRIDGE.name.lower() in res_url:
        logger.debug(f"{OP.PARSING.name} {res_url}")
        soup = make_a_soup(res_text)
        cambridge.parse_and_print(soup, res_url)
        c_print(f'\n#[#757575]{OP.FOUND.name} "{res_word}" from {DICT.CAMBRIDGE.name} in cache. You can add "-f -w" to fetch the {DICT.MERRIAM_WEBSTER.name} dictionary', justify="left")
    else:
        nodes = webster.parse_dict(res_text, True, res_url, False)
        webster.parse_and_print(nodes, res_url)
        c_print(f'\n#[#757575]{OP.FOUND.name} "{res_word}" from {DICT.MERRIAM_WEBSTER.name} in cache. You can add "-f" to fetch the {DICT.CAMBRIDGE.name} dictionary', justify="left")
    return True


def save(input_word, response_word, response_url, response_text):
    result = insert_into_table(input_word, response_word, response_url, response_text)
    if result[0]:
        logger.debug(f'{OP.CACHED.name} the search result of "{input_word}"')
    else:
        error_str = str(result[1])
        if "UNIQUE constraint" in error_str:
            if "response_word" in error_str:
                delete_word(response_word)
                insert_into_table(input_word, response_word, response_url, response_text)
                logger.debug(f'{OP.UPDATED.name} cache for "{input_word}" with the new search result\n')
            else:
                logger.debug(f'{OP.CANCELLED.name} caching "{input_word}", because it has been already cached before\n')
        else:
            logger.debug(f'{OP.CANCELLED.name} caching "{input_word}" - {error}\n')


def print_spellcheck(input_word, suggestions, dict_name, is_ch=False):
    is_cambridge = (dict_name == DICT.CAMBRIDGE.name)
    flip_dict = DICT.MERRIAM_WEBSTER.name if is_cambridge else DICT.CAMBRIDGE.name

    if is_tool("fzf"):
        spellcheck_notice = "NOT FOUND. Select to print the suggestion's meaning; [NUMBER] to switch to " + flip_dict + "; Input a new word; [ESC] to quit out."
        list_items_fzf(suggestions, "spell_check", spellcheck_notice, input_word, dict_name, is_ch)
    else:
        c_print(dict_name + (int(len(dict_name)/2))*" ", justify="center")
        list_items(suggestions, "spell_check")
        spellcheck_notice = "NOT FOUND. [NUMBER] to print the suggestion's meaning; [ENTER] to switch to " + flip_dict + "; Input a new word; [ANY OTHER KEY] to quit out: "
        c_print(spellcheck_notice, end="")
        key = input("")

        if (key.isnumeric() and (1 <= int(key) <= len(suggestions))):
            cambridge.search_cambridge(suggestions[int(key) - 1], False, is_ch) if is_cambridge else webster.search_webster(suggestions[int(key) - 1])
        elif len(key) > 1:
            cambridge.search_cambridge(key, False, is_ch) if is_cambridge else webster.search_webster(key)
        elif key == "":
            webster.search_webster(input_word) if is_cambridge else cambridge.search_cambridge(input_word, False, is_ch)
        else:
            sys.exit()


def print_word_per_line(index, word, extra=""):
    cols = os.get_terminal_size().columns
    word_len = len(word)

    if index % 2 == 0:
        print(f"\033[37;1;100m{index+1:6d}|{word}", end="")
        print(f"\033[37;1;100m{extra:>{cols-word_len-7}}\033[0m")
    else:
        c_print(f"#[bold #4A7D95]{index+1:6d}|{word}", end="")
        c_print(f"#[bold #4A7D95]{extra:>{cols-word_len-7}}")


def list_items(data, initiator: Optional[Initiator] = None):
    for index, entry in enumerate(data):
        if initiator == "cache_list":
            word = entry[0]
            dict_name = "CAMBRIDGE" if "cambridge" in entry[1] else "WEBSTER"
            print_word_per_line(index, word, dict_name)
        elif initiator == "wod_calendar":
            date_string = data[entry].split("/")[-1]
            date = ""
            for i, c in enumerate(date_string):
                if c.isnumeric():
                    date = date_string[i:]
                    break
            print_word_per_line(index, entry, date)
        else:
            print_word_per_line(index, entry, "")


def list_items_fzf(data, initiator: Optional[Initiator] = None, notice="", input_word=None, dict_name=None, is_ch=False):
    choices = ""
    if initiator == "cache_list":
        for i in data:
            choices = choices + i[0] + "\n"
    elif initiator == "wod_calendar":
        choices = "\n".join(data.keys())
    else:
        choices = "\n".join(data)
    choices = choices.strip("\n")

    p1 = subprocess.Popen(["echo", choices], stdout=subprocess.PIPE, text=True)
    p2 = subprocess.Popen(["fzf", "--layout=reverse", "--bind", "enter:accept-or-print-query", "--header", notice], stdin=p1.stdout, stdout=subprocess.PIPE, text=True)

    select_word = p2.communicate()[0].strip("\n")

    logger.debug(f"select_word by fzf is: {repr(select_word)}; returncode by fzf is: {p2.returncode}")

    is_cambridge = (dict_name == DICT.CAMBRIDGE.name)

    if len(select_word) == 1 and not select_word.isalpha() and initiator == "spell_check":
        webster.search_webster(input_word, req_url=None) if is_cambridge else cambridge.search_cambridge(input_word, is_ch=is_ch, req_url=None)
    elif p2.returncode == 0 and select_word != "":
        if initiator == "wod_calendar" and select_word in data.keys():
            url = webster.WEBSTER_BASE_URL + data[select_word]
            webster.get_wod_past(url)
        else:
            cambridge.search_cambridge(select_word, is_ch=is_ch, req_url=None) if is_cambridge else webster.search_webster(select_word, req_url=None)
    else:
        sys.exit()



import time
import cProfile
import pstats
import io
from urllib import parse
from functools import wraps
from bs4 import BeautifulSoup
from enum import Enum


class OP(Enum):
    FETCHING        = 1,
    PARSING         = 2,
    RETRY_FETCHING  = 3,
    RETRY_PARSING   = 4,
    PRINTING        = 5,
    FOUND           = 6,
    NOT_FOUND       = 7,
    CACHED          = 8,
    CANCELLED       = 9,
    DELETED         = 10,
    UPDATED         = 11


class DICT(Enum):
    CAMBRIDGE       = 1,
    MERRIAM_WEBSTER = 2


def make_a_soup(text):
    soup = BeautifulSoup(text, "lxml")
    return soup


def replace_all(string):
    return (
        string.replace("\n            (", "(")
        .replace("(\n    \t                ", "(")
        .replace("\n", " ")
        .replace("      \t                ", " ")
        .replace("\\'", "'")
        .replace("  ", " ")
        .replace("\xa0 ", "")
        .replace("[ ", "")
        .replace(" ]", "")
        .replace("[", "")
        .replace("]", "")
        .replace("A1", "")
        .replace("A2", "")
        .replace("B1", "")
        .replace("B2", "")
        .replace("C1", "")
        .replace("C2", "")
        .strip()
    )


def parse_response_url(url):
    """Return the unquoted Cambridge url for displaying and saving."""

    response_base_url = url.split("?")[0]
    return response_base_url


def get_request_url(url, input_word, dict):
    """Return the url formatted to request the web page."""

    if dict == DICT.CAMBRIDGE.name:
        if " " in input_word:
            query_word = input_word.replace(" ", "+").replace("/", "+")
            request_url = url + query_word
        else:
            request_url = url + input_word
        return request_url

    return url + parse.quote(input_word)


def decode_url(url):
    return parse.unquote(url)


def is_tool(name):
    """Check whether `name` is on $PATH and marked as executable."""

    from shutil import which
    return which(name) is not None


def profile(func):
    """A decorator that uses cProfile to profile a function"""

    def inner(*args, **kwargs):
        pr = cProfile.Profile()
        pr.enable()
        retval = func(*args, **kwargs)
        pr.disable()
        s = io.StringIO()
        sortby = "cumulative"
        ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
        ps.print_stats()
        print(s.getvalue())
        return retval

    return inner


def timer(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        start_at = time.time()
        f = func(*args, **kwargs)
        time_taken = round((time.time() - start_at), 2)
        print("\nTime taken: {} seconds".format(time_taken))
        return f

    return wrapper



import requests
import threading
import sys
import re
from ..console import c_print
from ..errors import ParsedNoneError, NoResultError, call_on_error
from ..log import logger
from ..utils import (
    make_a_soup,
    get_request_url,
    parse_response_url,
    replace_all,
    OP,
    DICT
)
from ..dicts import dicts


CAMBRIDGE_URL = "https://dictionary.cambridge.org"
CAMBRIDGE_EN_SEARCH_URL = CAMBRIDGE_URL + "/search/direct/?datasetsearch=english&q="
CAMBRIDGE_CN_SEARCH_URL = CAMBRIDGE_URL + "/search/direct/?datasetsearch=english-chinese-simplified&q="

CAMBRIDGE_SPELLCHECK_URL = CAMBRIDGE_URL + "/spellcheck/english/?q="
CAMBRIDGE_SPELLCHECK_URL_CN = CAMBRIDGE_URL + "/spellcheck/english-chinese-simplified/?q="

word_entry = ""
response_word = ""


def search_cambridge(input_word, is_fresh=False, is_ch=False, no_suggestions=False, req_url=None):
    if req_url is None:
        if is_ch:
            req_url = get_request_url(CAMBRIDGE_CN_SEARCH_URL, input_word, DICT.CAMBRIDGE.name)
        else:
            req_url = get_request_url(CAMBRIDGE_EN_SEARCH_URL, input_word, DICT.CAMBRIDGE.name)

    if not is_fresh:
        cached = dicts.cache_run(input_word, req_url)
        if not cached:
            fresh_run(req_url, input_word, is_ch, no_suggestions)
    else:
        fresh_run(req_url, input_word, is_ch, no_suggestions)


def fetch_cambridge(req_url, input_word, is_ch):
    with requests.Session() as session:
        session.trust_env = False # not to use proxy
        res = dicts.fetch(req_url, session)

        if "spellcheck" in res.url:
            logger.debug(f'{OP.NOT_FOUND.name} "{input_word}" in {DICT.CAMBRIDGE.name}')
            if is_ch:
                spell_req_url = get_request_url(CAMBRIDGE_SPELLCHECK_URL_CN, input_word, DICT.CAMBRIDGE.name)
            else:
                spell_req_url = get_request_url(CAMBRIDGE_SPELLCHECK_URL, input_word, DICT.CAMBRIDGE.name)

            spell_res = dicts.fetch(spell_req_url, session)
            spell_res_url = spell_res.url
            spell_res_text = spell_res.text
            return False, (spell_res_url, spell_res_text)

        else:
            res_url = parse_response_url(res.url)
            res_text = res.text

            logger.debug(f'{OP.FOUND.name} "{input_word}" in {DICT.CAMBRIDGE.name} at {res_url}')
            return True, (res_url, res_text)


def fresh_run(req_url, input_word, is_ch, no_suggestions=False):
    result = fetch_cambridge(req_url, input_word, is_ch)
    found = result[0]

    if found:
        res_url, res_text = result[1]
        soup = make_a_soup(res_text)

        global response_word
        response_word = parse_response_word(soup)
        global word_entry
        word_entry = input_word

        first_dict = parse_first_dict(res_url, soup)
        parse_and_print(first_dict, res_url, True)

    else:
        if no_suggestions:
            sys.exit(-1)
        else:
            spell_res_url, spell_res_text = result[1]

            logger.debug(f"{OP.PARSING.name} {spell_res_url}")
            soup = make_a_soup(spell_res_text)
            nodes = soup.find("div", "hfl-s lt2b lmt-10 lmb-25 lp-s_r-20")
            suggestions = []

            if not nodes:
                print(NoResultError(DICT.CAMBRIDGE.name))
                sys.exit()

            for ul in nodes.find_all("ul", "hul-u"):
                if "We have these words with similar spellings or pronunciations:" in ul.find_previous_sibling().text:
                    for i in ul.find_all("li"):
                        sug = replace_all(i.text)
                        suggestions.append(sug)

            logger.debug(f"{OP.PRINTING.name} the parsed result of {spell_res_url}")
            dicts.print_spellcheck(input_word, suggestions, DICT.CAMBRIDGE.name, is_ch)


def parse_and_print(first_dict, res_url, new_line=False):
    logger.debug(f"{OP.PRINTING.name} the parsed result of {res_url}")

    attempt = 0
    while True:
        try:
            blocks = first_dict.find_all("div", ["pr entry-body__el", "entry-body__el clrd js-share-holder", "pr idiom-block"])
        except AttributeError as e:
            attempt = call_on_error(e, res_url, attempt, OP.RETRY_PARSING.name)
            continue
        else:
            if len(blocks) != 0:
                for block in blocks:
                    parse_dict_head(block)
                    parse_dict_body(block)

                if new_line:
                    print()

                dicts.save(word_entry, response_word, res_url, str(first_dict))
                return
            else:
                print(NoResultError(DICT.CAMBRIDGE.name))
                sys.exit()


def parse_first_dict(res_url, soup):
    attempt = 0
    logger.debug(f"{OP.PARSING.name} {res_url}")

    while True:
        first_dict = soup.find("div", "pr dictionary") or soup.find("div", "pr di superentry")
        if first_dict is None:
            attempt = call_on_error(ParsedNoneError(DICT.CAMBRIDGE.name, res_url), res_url, attempt, OP.RETRY_PARSING.name)
            continue
        else:
            break

    return first_dict


def parse_response_word(soup):
    temp = soup.find("title").text.split("-")[0].strip()
    if "|" in temp:
        response_word = temp.split("|")[0].strip().lower()

    elif "in Simplified Chinese" in temp:
        response_word = temp.split("in Simplified Chinese")[0].strip().lower()
    else:
        response_word = temp.lower()

    return response_word


def parse_head_title(block):
    word = block.find("div", "di-title").text
    return word


def parse_head_info(block):
    info = block.find_all("span", ["pos dpos", "lab dlab", "v dv lmr-0"])

    if len(info) != 0:
        temp = [i.text for i in info]
        type = temp[0]
        text = " ".join(temp[1:])
        return (type, text)
    return None


def parse_head_type(head):
    anc = head.find("span", "anc-info-head danc-info-head")
    posgram = head.find("div", "posgram dpos-g hdib lmr-5")
    dpos = head.find("span", attrs={"title": "A word that describes an action, condition or experience."})

    w_type = ""
    if anc is not None:
        # e.g. "phrasal verb with sneak" superfluous
        # w_type = anc.text
        if dpos is not None:
            w_type += dpos.text
            dgram = dpos.find_next_sibling("span", "gram dgram")
            if dgram is not None:
                w_type += " " + dgram.text
        w_type = replace_all(w_type)
    elif posgram is not None:
        w_type = replace_all(posgram.text)
    return w_type


def print_pron(block, area):
    w_pron = block.find("span", "pron dpron")
    if w_pron is not None:
        w_pron_text = replace_all(w_pron.text).replace("/", "|")
        c_print(f"#[bold]{area} #[/bold]" + w_pron_text, end=" ")


def parse_head_pron(head):
    uk_dpron = head.find("span", "uk dpron-i")
    if uk_dpron is not None:
        print_pron(uk_dpron, "UK")

    us_dpron = head.find("span", "us dpron-i")
    if us_dpron is not None:
        print_pron(us_dpron, "US")


def parse_head_tense(block):
    sub_blocks = block.find_all("span", "inf-group dinfg")
    if len(sub_blocks) != 0:
        for index, sub in enumerate(sub_blocks):
            tenses = sub.find_all("b", "inf dinf")
            if len(tenses) != 0:
                for tense in tenses:
                    c_print("#[bold]" + tense.text + "#[/bold]", end=" ")
            if index == 0 and len(sub_blocks) > 1:
                print("|", end=" ")


def parse_head_domain(block):
    domain = replace_all(block.text)
    print(domain, end="  ")


def parse_head_var(head):
    var_dvar = head.find("span", "var dvar")
    if var_dvar is not None:
        w_var = replace_all(var_dvar.text)
        print(w_var, end="  ")

    var_dvar_sib = head.find_next_sibling("span", "var dvar")
    if var_dvar_sib is not None:
        w_var = replace_all(var_dvar_sib.text)
        print(w_var, end="  ")


def parse_head_spellvar(block):
    for i in block:
        spell_var = replace_all(i.text)
        print(spell_var, end="  ")


def parse_dict_head(block):
    head = block.find("div", "pos-header dpos-h")
    word = parse_head_title(block)
    info = parse_head_info(block)

    if head is not None:
        w_type = parse_head_type(head)

        if not word:
            word = parse_head_title(head)
        if w_type:
            w_type = w_type.replace(" or ", "/")
            c_print(f"\n#[bold blue]{word}#[/bold blue] #[bold yellow]{w_type}#[/bold yellow]")

        parse_head_pron(head)

        irreg = head.find("span", "irreg-infls dinfls")
        if irreg is not None:
            parse_head_tense(irreg)

        domain = head.find("span", "domain ddomain")
        if domain is not None:
            parse_head_domain(domain)

        parse_head_var(head)

        spellvar = head.find_all("span", "spellvar dspellvar")
        if len(spellvar) != 0:
            parse_head_spellvar(spellvar)

        print()

    else:
        c_print("#[bold blue]" + word)
        if info:
            print(f"{info[0]} {info[1]}")


# ----------Parse Dict Body----------
def parse_def_title(block):
    d_title = replace_all(block.find("h3", "dsense_h").text)
    c_print("#[red]" + "\n" + d_title.upper())

def parse_ptitle(block):
    p_title = block.find("span", "phrase-title dphrase-title").text
    p_info = block.find("span", "phrase-info dphrase-info")

    if p_info is not None:
        phrase_info = replace_all(p_info.text)
        print(f"\033[34;1m  {p_title}\033[0m \033[33;1m{phrase_info}\033[0m")
    else:
        print(f"\033[34;1m  {p_title}\033[0m")


def parse_def_info(def_block):
    """ def info tags like 'B1 [ C or U ]' """

    def_info = replace_all(def_block.find("span", "def-info ddef-info").text).replace(" or ", "/")
    return def_info


def print_meaning_lan(meaning_lan):
    if meaning_lan is not None:
        meaning_lan_words = meaning_lan.text.replace(";", "；").replace(",", "，")
        if not meaning_lan_words.startswith("（"):
            print(" ", end="")
        print("\033[34m" + meaning_lan_words + "\033[0m")
    else:
        print()


def print_meaning(meaning_b, usage_b, is_pmeaning):
    if is_pmeaning:
        print("  ", end="")

    if usage_b is not None:
        usage = replace_all(usage_b.text)
        meaning_words = replace_all(meaning_b.text).split(usage)[-1].replace(":", "")
        print("\033[34;1m: \033[0m" + "[" + usage + "]" + "\033[34m" + meaning_words + "\033[0m", end="")
    else:
        meaning_words = replace_all(meaning_b.text).replace(":", "")
        print("\033[34;1m: \033[0m" + "\033[34m" + meaning_words + "\033[0m", end="")


def parse_meaning(def_block, is_pmeaning=False):
    meaning_b = def_block.find("div", "def ddef_d db")
    usage_b = meaning_b.find("span", "lab dlab")

    print_meaning(meaning_b, usage_b, is_pmeaning)

    def_info = parse_def_info(def_block)
    if def_info:
        if not def_info.startswith("("):
            print(" [" + def_info + "]", end="")
        else:
            print(" " + def_info, end="")

    # Print the meaning's specific language translation if any
    meaning_lan = def_block.find("span", "trans dtrans dtrans-se break-cj")
    print_meaning_lan(meaning_lan)


def print_example_tag(tag_block):
    if tag_block is None:
        return

    tag = tag_block.text
    tag = replace_all(tag)
    print("[" + tag + "]", end= " ")


def parse_example(def_block, is_pexample=False):
    es = def_block.find_all("div", "examp dexamp")
    if len(es) == 0:
        return

    for e in es:
        result = e.find("span", "eg deg")
        if result is not None:
            example = replace_all(result.text)
        else:
            continue

        if is_pexample:
            print("  ", end="")

        example_lan = e.find("span", "trans dtrans dtrans-se hdb break-cj")
        if example_lan is not None:
            example_lan_sent = " " + example_lan.text.replace("。", "")
        else:
            example_lan_sent = ""

        c_print("#[blue]" + "|" + "#[/blue]", end="")

        dlab = e.find("span", "lab dlab")
        print_example_tag(dlab)

        dgram = e.find("span", "gram dgram")
        print_example_tag(dgram)

        dlu = e.find("span", "lu dlu")
        print_example_tag(dlu)

        c_print(f"#[#757575]{example}{example_lan_sent}#[/#757575]")


def print_tag(node):
    for i in node.find_all("span"):
        has_class = i.has_attr("class")
        parent = i.find_parent()
        parent_has_class = parent.has_attr("class")

        # bs4 treats `class="x-h dx-h"` as a dict, key is 'class', value is ['x-h', 'dx-h']
        # which is counterintuitive, especially with the whole class value as a list.

        if has_class and i.attrs["class"] == ['x-h', 'dx-h']:
            c_print("#[#757575]" + "  • " + i.text, end="")
        elif parent_has_class and parent.attrs["class"] == ['x-h', 'dx-h']:
            continue
        elif has_class and i.attrs["class"] == ['x-lab', 'dx-lab']:
            c_print("#[#757575]" + " [" + i.text + "]", end="")
        elif parent_has_class and parent.attrs["class"] == ['x-lab', 'dx-lab']:
            continue
        elif has_class and i.attrs["class"] == ['x-pos', 'dx-pos']:
            print(" " + i.text, end="")
        else:
            c_print("#[#757575]" + " " + i.text, end="")
    print()


def print_synonym(block):
    s_title = block.strong.text.upper()
    c_print("#[bold #757575]" + "\n  " + s_title)

    for item in block.find_all("div", ["item lc lc1 lpb-10 lpr-10", "item lc lc1 lc-xs6-12 lpb-10 lpr-10"]):
        print_tag(item)


def parse_synonym(block):
    s_block = block.find("div", re.compile("xref synonyms? hax dxref-w( lmt-25)?"))
    if s_block is not None:
        print_synonym(s_block)
    elif "synonym" in block.attrs["class"]:
        print_synonym(block)


def parse_see_also(def_block):
    see_also_block = def_block.find("div", re.compile("xref see_also hax dxref-w( lmt-25)?"))

    if see_also_block is not None:
        see_also = see_also_block.strong.text.upper()
        c_print("#[bold #757575]" + "\n  " + see_also)
        print_tag(see_also_block)


def parse_compare(def_block):
    compare_block = def_block.find("div", re.compile("xref compare hax dxref-w( lmt-25)?"))

    if compare_block is not None:
        compare = compare_block.strong.text.upper()
        c_print("#[bold #757575]" + "\n  " + compare)
        for item in compare_block.find_all("div", ["item lc lc1 lpb-10 lpr-10", "item lc lc1 lc-xs6-12 lpb-10 lpr-10"]):
            print_tag(item)


def parse_usage_note(def_block):
    usage_block = def_block.find("div", "usagenote dusagenote daccord")

    if usage_block is not None:
        usagenote = usage_block.h5.text
        c_print("#[bold #757575]" + "\n  " + usagenote)
        for item in usage_block.find_all("li", "text"):
            item = item.text
            c_print("#[#757575]" + "    " + item)


def parse_def(def_block):
    if "phrase-body" in def_block.parent.attrs["class"]:
        parse_meaning(def_block, True)
        parse_example(def_block, True)
    else:
        parse_meaning(def_block)
        parse_example(def_block)

    parse_synonym(def_block)
    parse_see_also(def_block)
    parse_compare(def_block)
    parse_usage_note(def_block)

def parse_idiom(block):
    idiom_block = block.find("div", re.compile("xref idioms? hax dxref-w lmt-25 lmb-25"))

    if idiom_block is not None:
        idiom_title = idiom_block.h3.text.upper()
        c_print("#[bold #757575]" + "\n" + idiom_title)
        for item in idiom_block.find_all("div", ["item lc lc1 lpb-10 lpr-10", "item lc lc1 lc-xs6-12 lpb-10 lpr-10"]):
            print_tag(item)


def parse_sole_idiom(block):
    idiom_sole_meaning = block.find("div", "def ddef_d db")

    if idiom_sole_meaning is not None:
        print("\033[34m" + idiom_sole_meaning.text + "\033[0m")
    parse_example(block)
    parse_see_also(block)


def parse_phrasal_verb(block):
    pv_block = block.find("div", re.compile("xref phrasal_verbs? hax dxref-w lmt-25 lmb-25"))

    if pv_block is not None:
        pv_title = pv_block.h3.text.upper()
        c_print("#[bold #757575]" + "\n" + pv_title)
        for item in pv_block.find_all("div", ["item lc lc1 lc-xs6-12 lpb-10 lpr-10", "item lc lc1 lpb-10 lpr-10"]):
            print_tag(item)


def parse_dict_body(block):
    subblocks = block.find_all("div", ["pr dsense", "pr dsense dsense-noh"])

    if len(subblocks) != 0:
        for subblock in subblocks:
            # Comment out because h3 block seems superfluous.
            # if subblock.find("h3", "dsense_h"):
            #     parse_def_title(subblock)

            sense = subblock.find("div", "sense-body dsense_b")
            if sense is not None:
                for child in sense.children:
                    try:
                        attr = child.attrs["class"]
                        if attr and attr == ["def-block", "ddef_block"]:
                            parse_def(child)

                        if attr and "synonym" in attr:
                            parse_synonym(child)

                        if attr and (attr == ["pr", "phrase-block", "dphrase-block", "lmb-25"] or attr == ["pr", "phrase-block", "dphrase-block"]):
                            parse_ptitle(child)

                            for i in child.find_all("div", "def-block ddef_block"):
                                parse_def(i)
                    except Exception:
                        pass

    else:
        idiom_sole_block = block.find("div", "idiom-block")
        if idiom_sole_block is not None:
            parse_sole_idiom(idiom_sole_block)

    parse_idiom(block)
    parse_phrasal_verb(block)


def parse_dict_name(first_dict):
    if first_dict.small is not None:
        dict_info = replace_all(first_dict.small.text).strip("(").strip(")")
        dict_name = dict_info.split("©")[0]
        dict_name = dict_name.split("the")[-1]
    else:
        dict_name="Cambridge Dictionary"
    c_print(f"#[#757575]{dict_name}", justify="right")



def main():
    try:
        args = parse_args()
        if args is not None:
            args.func(args)

        if con:
            con.close()

    except KeyboardInterrupt:
        print("\nStopped by user")


if __name__ == "__main__":
    import os
    import sys

    sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

    from cambridge.args import parse_args
    from cambridge.cache import con

else:
    from .args import parse_args
    from .cache import con

main()
